from scipy.signal import fftconvolve
import numpy as np
import astropy.units as U
from astropy.io import fits
from astropy import __version__ as astropy_version
from datetime import datetime
from itertools import product


class Martini():
    """
    Used for the creation of synthetic HI data cubes from simulation data.

    Usual use of martini involves first creating instances of classes from each
    of the required and optional sub-modules, then creating a Martini with
    these instances as arguments. The object can then be used to create
    synthetic observations, usually by calling 'insert_source_in_cube',
    (optionally) 'add_noise', (optionally) 'convolve_beam' and 'write_fits' in
    order.

    Parameters
    ----------
    source : an instance of a class derived from martini.source._BaseSource
        A description of the HI emitting object, including position, geometry
        and an interface to the simulation data (SPH particle masses,
        positions, etc.). Sources leveraging the simobj package for reading
        simulation data (github.com/kyleaoman/simobj) and a few test sources
        (e.g. single particle) are provided, creation of customized sources,
        for instance to leverage other interfaces to simulation data, is
        straightforward. See sub-module documentation.

    datacube : martini.DataCube instance
        A description of the datacube to create, including pixels, channels,
        sky position. See sub-module documentation.

    beam : (optional) an instance of a class derived from beams._BaseBeam
        A description of the beam for the simulated telescope. Given a
        description, either mathematical or as an image, the creation of a
        custom beam is straightforward. See sub-module documentation.

    noise : (optional) an instance of a class derived from noise._BaseNoise
        A description of the simulated noise. A simple Gaussian noise model is
        provided; implementation of other noise models is straightforward. See
        sub-module documentation.

    sph_kernel : an instance of a class derived from sph_kernels._BaseSPHKernel
        A description of the SPH smoothing kernel. The Wendland C2 kernel (used
        in EAGLE), and a point-like Dirac-delta kernel are implemented,
        implementation of other kernels is straightforward. See sub-module
        documentation.

    spectral_model : an instance of a class derived from
                     spectral_models._BaseSpectrum
        A description of the HI line produced by a particle of given
        properties. A Dirac-delta spectrum, and both fixed-width and
        temperature-dependent Gaussian line models are provided; implementing
        other models is straightforward. See sub-module documentation.

    logtag : string
        String to prepend to standard output messages.

    Returns
    -------
    out : Martini
        A Martini object configured with the specified sub-modules.

    See Also
    --------
    martini.sources
    martini.DataCube
    martini.beams
    martini.noise
    martini.sph_kernels
    martini.spectral_models

    Examples
    --------
    The following example illustrates basic use of martini, using a (very!)
    crude model of a gas disk. This example can be run by doing
    'from martini import demo; demo()'.

    from martini import Martini, DataCube
    from martini.beams import GaussianBeam
    from martini.noise import GaussianNoise
    from martini.spectral_models import GaussianSpectrum
    from martini.sph_kernels import DiracDeltaKernel
    from martini.sources import SPHSource
    import astropy.units as U
    import numpy as np

    # ------make a toy galaxy----------
    N = 1000
    phi = np.random.rand(N) * 2 * np.pi
    r = []
    for L in np.random.rand(N):
        def f(r):
            return L - .5 * (2 - np.exp(-r) * (np.power(r, 2) + 2 * r + 2))
        r.append(fsolve(f, 1.)[0])
    r = np.array(r)
    # exponential disk
    r *= 3 / np.sort(r)[N // 2]
    z = -np.log(np.random.rand(N))
    # exponential scale height
    z *= .5 / np.sort(z)[N // 2] * np.sign(np.random.rand(N) - .5)
    x = r * np.cos(phi)
    y = r * np.sin(phi)
    xyz_g = np.vstack((x, y, z)) * U.kpc
    # linear rotation curve
    vphi = 100 * r / 6.
    vx = -vphi * np.sin(phi)
    vy = vphi * np.cos(phi)
    # small pure random z velocities
    vz = (np.random.rand(N) * 2. - 1.) * 5
    vxyz_g = np.vstack((vx, vy, vz)) * U.km * U.s ** -1
    T_g = np.ones(N) * 8E3 * U.K
    mHI_g = np.ones(N) / N * 5.E9 * U.Msun
    # ~mean interparticle spacing smoothing
    hsm_g = np.ones(N) * 2 / np.sqrt(N) * U.kpc
    # ---------------------------------

    source = SPHSource(
        distance=5. * U.Mpc,
        rotation={'L_coords': (60. * U.deg, 0. * U.deg)},
        ra=0. * U.deg,
        dec=0. * U.deg,
        h=.7,
        T_g=T_g,
        mHI_g=mHI_g,
        xyz_g=xyz_g,
        vxyz_g=vxyz_g,
        hsm_g=hsm_g
    )

    datacube = DataCube(
        n_px_x=128,
        n_px_y=128,
        n_channels=32,
        px_size=10. * U.arcsec,
        channel_width=10. * U.km * U.s ** -1,
        velocity_centre=source.vsys
    )

    beam = GaussianBeam(
        bmaj=30. * U.arcsec,
        bmin=30. * U.arcsec,
        bpa=0. * U.deg,
        truncate = 4.
    )

    noise = GaussianNoise(
        rms=3.E-4 * U.Jy * U.arcsec ** -2
    )

    spectral_model = GaussianSpectrum(
        sigma=7 * U.km * U.s ** -1
    )

    sph_kernel = DiracDeltaKernel()

    M = Martini(
        source=source,
        datacube=datacube,
        beam=beam,
        noise=noise,
        spectral_model=spectral_model,
        sph_kernel=sph_kernel
    )

    M.insert_source_in_cube()
    M.add_noise()
    M.convolve_beam()
    M.write_beam_fits('testbeam.fits', channels='velocity')
    M.write_fits('testcube.fits', channels='velocity')
    """

    def __init__(self,
                 source=None,
                 datacube=None,
                 beam=None,
                 noise=None,
                 sph_kernel=None,
                 spectral_model=None,
                 logtag=''):
        self.source = source
        self.datacube = datacube
        self.beam = beam
        self.noise = noise
        self.sph_kernel = sph_kernel
        self.spectral_model = spectral_model
        self.logtag = logtag

        if self.beam is not None:
            self.beam.init_kernel(self.datacube)
            self.datacube.add_pad(self.beam.needs_pad())

        self._prune_source()

        self.spectral_model.init_spectra(self.source, self.datacube)

        return

    def convolve_beam(self):
        """
        Convolve the beam and DataCube.
        """

        unit = self.datacube._array.unit
        for spatial_slice in self.datacube.spatial_slices():
            # use a view [...] to force in-place modification
            spatial_slice[...] = fftconvolve(
                spatial_slice, self.beam.kernel, mode='same') * unit
        self.datacube.drop_pad()
        self.datacube._array = self.datacube._array.to(
            U.Jy * U.beam**-1, equivalencies=[self.beam.arcsec_to_beam])
        return

    def add_noise(self):
        """
        Insert noise into the DataCube.
        """

        self.datacube._array = \
            self.datacube._array + self.noise.generate(self.datacube)
        return

    def _prune_source(self):
        """
        Determines which particles cannot contribute to the DataCube and
        removes them to speed up calculation. Assumes the kernel is 0 at
        distances greater than the SPH smoothing length.
        """

        # pixels indexed from 0 (not like in FITS!) for better use with numpy
        origin = 0
        skycoords = self.source.sky_coordinates
        particle_coords = np.vstack(
            self.datacube.wcs.sub(3).wcs_world2pix(
                skycoords.ra.to(self.datacube.units[0]),
                skycoords.dec.to(self.datacube.units[1]),
                skycoords.radial_velocity.to(self.datacube.units[2]),
                origin)) * U.pix
        # could use a function bound to source which returns the size of the
        # kernel, in case this isn't equal to the smoothing length for some
        # kernel
        sm_length = np.arctan(
            self.source.hsm_g / self.source.sky_coordinates.distance).to(
                U.pix, U.pixel_scale(self.datacube.px_size / U.pix))
        sm_range = np.ceil(sm_length).astype(int)
        spectrum_half_width = self.spectral_model.half_width(self.source) / \
            self.datacube.channel_width
        reject_conditions = (
            (particle_coords[:2] + sm_range[np.newaxis] <
             0 * U.pix).any(axis=0),
            particle_coords[0] - sm_range >
            (self.datacube.n_px_x + self.datacube.padx * 2) * U.pix,
            particle_coords[1] - sm_range >
            (self.datacube.n_px_y + self.datacube.pady * 2) * U.pix,
            particle_coords[2] + 4 * spectrum_half_width * U.pix < 0 * U.pix,
            particle_coords[2] - 4 * spectrum_half_width * U.pix >
            self.datacube.n_channels * U.pix,
        )
        reject_mask = np.zeros(particle_coords[0].shape)
        for condition in reject_conditions:
            reject_mask = np.logical_or(reject_mask, condition)
        self.source.apply_mask(np.logical_not(reject_mask))
        return

    def insert_source_in_cube(self, skip_validation=False):
        """
        Populates the DataCube with flux from the particles in the source.

        Parameters
        ----------
        skip_validation : bool
            SPH kernel interpolation onto the DataCube is approximated for
            increased speed. For some combinations of pixel size, distance
            and SPH smoothing length, the approximation may break down. The
            kernel class will check whether this will occur and raise a
            RuntimeError if so. This validation can be skipped (at the cost
            of accuracy!) by setting this parameter True.
        """

        # pixels indexed from 0 (not like in FITS!) for better use with numpy
        origin = 0
        skycoords = self.source.sky_coordinates
        particle_coords = np.vstack(
            self.datacube.wcs.sub(3).wcs_world2pix(
                skycoords.ra.to(self.datacube.units[0]),
                skycoords.dec.to(self.datacube.units[1]),
                skycoords.radial_velocity.to(self.datacube.units[2]),
                origin)) * U.pix
        sm_length = np.arctan(
            self.source.hsm_g / self.source.sky_coordinates.distance).to(
                U.pix, U.pixel_scale(self.datacube.px_size / U.pix))
        if not skip_validation:
            self.sph_kernel.validate(sm_length)
        sm_range = np.ceil(sm_length).astype(int)

        # pixel iteration
        ij_pxs = list(
            product(
                np.arange(self.datacube._array.shape[0]),
                np.arange(self.datacube._array.shape[1])))
        print(
            '  ' + self.logtag + '  [columns: {0:.0f}, rows: {1:.0f}]'.format(
                self.datacube._array.shape[0], self.datacube._array.shape[1]))
        for ij_px in ij_pxs:
            ij = np.array(ij_px)[..., np.newaxis] * U.pix
            if (ij[1, 0].value == 0) and (ij[0, 0].value % 100 == 0):
                print('  ' + self.logtag +
                      '  [row {:.0f}]'.format(ij[0, 0].value))
            mask = (np.abs(ij - particle_coords[:2]) <= sm_range).all(axis=0)
            weights = self.sph_kernel.px_weight(particle_coords[:2, mask] - ij,
                                                sm_length[mask])
            self.datacube._array[ij_px[0], ij_px[1], :, 0] = \
                (self.spectral_model.spectra[mask] *
                 weights[..., np.newaxis]).sum(axis=-2)

        self.datacube._array = \
            self.datacube._array / np.power(self.datacube.px_size / U.pix, 2)
        return

    def write_fits(self, filename, channels='frequency', overwrite=True):
        """
        Output the DataCube to a FITS-format file.

        Parameters
        ----------
        filename : string
            Name of the file to write. '.fits' will be appended if not already
            present.

        channels : 'frequency' (default), or 'velocity'
            Type of units used along the spectral axis in output file.

        overwrite: bool
            Whether to allow overwriting existing files, note that the default
            is True.
        """

        self.datacube.drop_pad()
        if channels == 'frequency':
            self.datacube.freq_channels()
        elif channels == 'velocity':
            pass
        else:
            raise ValueError("Martini.write_fits: Unknown 'channels' value "
                             "(use 'frequency' or 'velocity'.")

        filename = filename if filename[-5:] == '.fits' else filename + '.fits'

        wcs_header = self.datacube.wcs.to_header()
        wcs_header.rename_keyword('WCSAXES', 'NAXIS')

        header = fits.Header()
        header.append(('SIMPLE', 'T'))
        header.append(('BITPIX', 16))
        header.append(('NAXIS', wcs_header['NAXIS']))
        header.append(('NAXIS1', self.datacube.n_px_x))
        header.append(('NAXIS2', self.datacube.n_px_y))
        header.append(('NAXIS3', self.datacube.n_channels))
        header.append(('NAXIS4', 1))
        header.append(('EXTEND', 'T'))
        header.append(('CDELT1', wcs_header['CDELT1']))
        header.append(('CRPIX1', wcs_header['CRPIX1']))
        header.append(('CRVAL1', wcs_header['CRVAL1']))
        header.append(('CTYPE1', wcs_header['CTYPE1']))
        header.append(('CUNIT1', wcs_header['CUNIT1']))
        header.append(('CDELT2', wcs_header['CDELT2']))
        header.append(('CRPIX2', wcs_header['CRPIX2']))
        header.append(('CRVAL2', wcs_header['CRVAL2']))
        header.append(('CTYPE2', wcs_header['CTYPE2']))
        header.append(('CUNIT2', wcs_header['CUNIT2']))
        header.append(('CDELT3', wcs_header['CDELT3']))
        header.append(('CRPIX3', wcs_header['CRPIX3']))
        header.append(('CRVAL3', wcs_header['CRVAL3']))
        header.append(('CTYPE3', wcs_header['CTYPE3']))
        header.append(('CUNIT3', wcs_header['CUNIT3']))
        header.append(('CDELT4', wcs_header['CDELT4']))
        header.append(('CRPIX4', wcs_header['CRPIX4']))
        header.append(('CRVAL4', wcs_header['CRVAL4']))
        header.append(('CTYPE4', wcs_header['CTYPE4']))
        header.append(('CUNIT4', 'PAR'))
        header.append(('EPOCH', 2000))
        header.append(('INSTRUME', 'WSRT', 'MARTINI Synthetic'))
        # header.append(('BLANK', -32768)) #only for integer data
        header.append(('BSCALE', 1.0))
        header.append(('BZERO', 0.0))
        header.append(('DATAMAX', np.max(self.datacube._array.value)))
        header.append(('DATAMIN', np.min(self.datacube._array.value)))
        header.append(('ORIGIN', 'astropy v' + astropy_version))
        # long names break fits format, don't let the user set this
        header.append(('OBJECT', 'MOCK'))
        if self.beam is not None:
            header.append(('BPA', self.beam.bpa.to(U.deg).value))
        header.append(('OBSERVER', 'K. Oman'))
        # header.append(('NITERS', ???))
        # header.append(('RMS', ???))
        # header.append(('LWIDTH', ???))
        # header.append(('LSTEP', ???))
        header.append(('BUNIT', str(self.datacube._array.unit).replace(
            ' ', '')))
        # header.append(('PCDEC', ???))
        # header.append(('LSTART', ???))
        header.append(('DATE-OBS', datetime.utcnow().isoformat()[:-5]))
        # header.append(('LTYPE', ???))
        # header.append(('PCRA', ???))
        # header.append(('CELLSCAL', ???))
        if self.beam is not None:
            header.append(('BMAJ', self.beam.bmaj.to(U.deg).value))
            header.append(('BMIN', self.beam.bmin.to(U.deg).value))
        header.append(('BTYPE', 'Intensity'))
        header.append(('SPECSYS', wcs_header['SPECSYS']))

        # flip axes to write
        hdu = fits.PrimaryHDU(header=header, data=self.datacube._array.value.T)
        hdu.writeto(filename, overwrite=overwrite)

        if channels == 'frequency':
            self.datacube.velocity_channels()
        return

    def write_beam_fits(self, filename, channels='frequency', overwrite=True):
        """
        Output the beam to a FITS-format file.

        The beam is written to file, with pixel sizes, coordinate system, etc.
        similar to those used for the DataCube.

        Parameters
        ----------
        filename : string
            Name of the file to write. '.fits' will be appended if not already
            present.

        channels : 'frequency' (default), or 'velocity'
            Type of units used along the spectral axis in output file.

        overwrite: bool
            Whether to allow overwriting existing files, note that the default
            is True.

        Raises
        ------
        ValueError
            If Martini was initialized without a beam.
        """

        if self.beam is None:
            raise ValueError("Martini.write_beam_fits: Called with beam set "
                             "to 'None'.")

        if channels == 'frequency':
            self.datacube.freq_channels()
        elif channels == 'velocity':
            pass
        else:
            raise ValueError("Martini.write_beam_fits: Unknown 'channels' "
                             "value (use 'frequency' or 'velocity'.")

        filename = filename if filename[-5:] == '.fits' else filename + '.fits'

        wcs_header = self.datacube.wcs.to_header()

        header = fits.Header()
        header.append(('SIMPLE', 'T'))
        header.append(('BITPIX', 16))
        # header.append(('NAXIS', self.beam.kernel.ndim))
        header.append(('NAXIS', 3))
        header.append(('NAXIS1', self.beam.kernel.shape[0]))
        header.append(('NAXIS2', self.beam.kernel.shape[1]))
        header.append(('NAXIS3', 1))
        header.append(('EXTEND', 'T'))
        header.append(('BSCALE', 1.0))
        header.append(('BZERO', 0.0))
        # this is 1/arcsec^2, is this right?
        header.append(('BUNIT', self.beam.kernel.unit.to_string('fits')))
        header.append(('CRPIX1', self.beam.kernel.shape[0] // 2 + 1))
        header.append(('CDELT1', wcs_header['CDELT1']))
        header.append(('CRVAL1', wcs_header['CRVAL1']))
        header.append(('CTYPE1', wcs_header['CTYPE1']))
        header.append(('CUNIT1', wcs_header['CUNIT1']))
        header.append(('CRPIX2', self.beam.kernel.shape[1] // 2 + 1))
        header.append(('CDELT2', wcs_header['CDELT2']))
        header.append(('CRVAL2', wcs_header['CRVAL2']))
        header.append(('CTYPE2', wcs_header['CTYPE2']))
        header.append(('CUNIT2', wcs_header['CUNIT2']))
        header.append(('CRPIX3', 1))
        header.append(('CDELT3', wcs_header['CDELT3']))
        header.append(('CRVAL3', wcs_header['CRVAL3']))
        header.append(('CTYPE3', wcs_header['CTYPE3']))
        header.append(('CUNIT3', wcs_header['CUNIT3']))
        header.append(('SPECSYS', wcs_header['SPECSYS']))
        header.append(('BMAJ', self.beam.bmaj.to(U.deg).value))
        header.append(('BMIN', self.beam.bmin.to(U.deg).value))
        header.append(('BPA', self.beam.bpa.to(U.deg).value))
        header.append(('BTYPE', 'beam    '))
        header.append(('EPOCH', 2000))
        header.append(('OBSERVER', 'K. Oman'))
        # long names break fits format
        header.append(('OBJECT', 'MOCKBEAM'))
        header.append(('INSTRUME', 'WSRT', 'MARTINI Synthetic'))
        header.append(('DATAMAX', np.max(self.beam.kernel.value)))
        header.append(('DATAMIN', np.min(self.beam.kernel.value)))
        header.append(('ORIGIN', 'astropy v' + astropy_version))

        # flip axes to write
        hdu = fits.PrimaryHDU(
            header=header, data=self.beam.kernel.value[..., np.newaxis].T)
        hdu.writeto(filename, overwrite=True)

        if channels == 'frequency':
            self.datacube.velocity_channels()
        return



"""
Mock APERTIF-like Radio Telescope Interferometry of the Neutral ISM

MARTINI is a modular package for the creation of synthetic resolved HI line
observations (data cubes) of smoothed-particle hydrodynamics simulations of
galaxies. The various aspects of the mock-observing process are divided
logically into sub-modules handling the data cube, source, beam, noise,
spectral model and SPH kernel. MARTINI is object-oriented: each sub-module
provides a class (or classes) which can be configured as desired. For most
sub-modules, base classes are provided to allow for straightforward
customization. Instances of each sub-module class are then given as parameters
to the Martini class. A mock observation is then constructed by calling a
handful of functions to execute the desired steps in the mock-observing
process.

The package is functional and (an old version) has been used in a forthcoming
paper (https://ui.adsabs.harvard.edu/#abs/2017arXiv170607478O/abstract). It is
still undergoing active development and is not quite yet at the point of a
stable release version. Things will change, bugs will happen. Any feedback is
greatly appreciated.

See the help for martini.Martini for an example script to configure MARTINI
and create a datacube. This example can be run by doing:

from martini import demo
demo()

If your use of MARTINI leads to a publication, please acknowledge this and link
to the github page, ideally specifying the version used (git commit ID).

Suport available via koman@astro.rug.nl.
"""

from ._martini import Martini
from ._datacube import DataCube
from ._demo import demo



import numpy as np
import astropy.units as U
from astropy import wcs

HIfreq = 1.420405751E9 * U.Hz


class DataCube():
    """
    Handles creation and management of the data cube itself.

    Basic usage simply involves initializing with the parameters listed below.
    More advanced usage might arise if designing custom classes for other sub-
    modules, especially beams.

    Parameters
    ----------
    n_px_x : int
        Pixel count along the x (RA) axis. Even integers strongly preferred.

    n_px_y : int
        Pixel count along the y (Dec) axis. Even integers strongly preferred.

    n_channels : int
        Number of channels along the spectral axis.

    px_size : astropy.units.Quantity, with dimensions of angle
        Angular scale of one pixel.

    channel_width : astropy.units.Quantity, with dimensions of velocity
        Step size along the spectral axis. Must be provided as a velocity.

    velocity_centre : astropy.units.Quantity, with dimensions of velocity
        Velocity of the central channel along the spectral axis.

    ra : astropy.units.Quantity, with dimensions of angle
        Right ascension of the cube centroid.

    dec : astropy.units.Quantity, with dimensions of angle
        Declination of the cube centroid.

    Returns
    -------
    out : DataCube
        An appropriately configured DataCube object.

    Examples
    --------
    TODO

    """

    def __init__(self,
                 n_px_x=256,
                 n_px_y=256,
                 n_channels=64,
                 px_size=15. * U.arcsec,
                 channel_width=4. * U.km * U.s**-1,
                 velocity_centre=0. * U.km * U.s**-1,
                 ra=0. * U.deg,
                 dec=0. * U.deg):

        datacube_unit = U.Jy * U.pix**-2
        self._array = np.zeros((n_px_x, n_px_y, n_channels, 1)) * datacube_unit
        self.n_px_x, self.n_px_y, self.n_channels = n_px_x, n_px_y, n_channels
        self.px_size = px_size
        self.channel_width = channel_width
        self.velocity_centre = velocity_centre
        self.ra = ra
        self.dec = dec
        self.padx = 0
        self.pady = 0
        self.wcs = wcs.WCS(naxis=3)
        self.wcs.wcs.crpix = [
            self.n_px_x / 2. + .5, self.n_px_y / 2. + .5, self.n_channels // 2
        ]
        self.units = [U.deg, U.deg, U.m * U.s**-1]
        self.wcs.wcs.cunit = [unit.to_string('fits') for unit in self.units]
        self.wcs.wcs.cdelt = [
            -self.px_size.to(self.units[0]).value,
            self.px_size.to(self.units[1]).value,
            self.channel_width.to(self.units[2]).value
        ]
        self.wcs.wcs.crval = [
            self.ra.to(self.units[0]).value,
            self.dec.to(self.units[1]).value,
            self.velocity_centre.to(self.units[2]).value
        ]
        self.wcs.wcs.ctype = ['RA---TAN', 'DEC--TAN', 'VELO-OBS']
        self.wcs = wcs.utils.add_stokes_axis_to_wcs(self.wcs,
                                                    self.wcs.wcs.naxis)
        self._channel_mids()
        self._channel_edges()
        self._freq_channel_mode = False

        return

    def _channel_mids(self):
        """
        Calculate the centres of the channels from the coordinate system.
        """
        self.channel_mids = self.wcs.wcs_pix2world(
            np.zeros(self.n_channels), np.zeros(self.n_channels),
            np.arange(self.n_channels), np.zeros(self.n_channels),
            0)[2] * self.units[2]
        return

    def _channel_edges(self):
        """
        Calculate the edges of the channels from the coordinate system.
        """
        self.channel_edges = self.wcs.wcs_pix2world(
            np.zeros(self.n_channels + 1), np.zeros(self.n_channels + 1),
            np.arange(self.n_channels + 1) - .5, np.zeros(self.n_channels + 1),
            0)[2] * self.units[2]
        return

    def spatial_slices(self):
        """
        Return an iterator over the spatial 'slices' of the cube.

        Returns
        -------
        out : iterator
            Iterator over the spatial 'slices' of the cube.
        """
        return iter(self._array[..., 0].transpose((2, 0, 1)))

    def spectra(self):
        """
        Return an iterator over the spectra (one in each spatial pixel).

        Returns
        -------
        out : iterator
            Iterator over the spectra (one in each spatial pixel).
        """
        return iter(self._array[..., 0].reshape(self.n_px_x * self.n_px_y,
                                                self.n_channels))

    def freq_channels(self):
        """
        Convert spectral axis to frequency units.
        """
        if self._freq_channel_mode:
            return
        else:

            def convert_to_Hz(q):
                return q.to(U.Hz, equivalencies=U.doppler_radio(HIfreq))
            self.wcs.wcs.cdelt[2] = \
                np.abs(convert_to_Hz(self.wcs.wcs.cdelt[2] * self.units[2])
                       - convert_to_Hz(0. * self.units[2])).value
            self.wcs.wcs.crval[2] = \
                convert_to_Hz(self.wcs.wcs.crval[2] * self.units[2]).value
            self.wcs.wcs.ctype[2] = 'FREQ-OBS'
            self.units[2] = U.Hz
            self.wcs.wcs.cunit[2] = self.units[2].to_string('fits')
            self._channel_mids()
            self._channel_edges()
            self._freq_channel_mode = True
            return

    def velocity_channels(self):
        """
        Convert spectral axis (back) to velocity units.
        """
        if not self._freq_channel_mode:
            return
        else:

            def convert_to_ms(q):
                return q.to(U.m / U.s, equivalencies=U.doppler_radio(HIfreq))
            self.wcs.wcs.cdelt[2] = \
                np.abs(convert_to_ms(self.wcs.wcs.cdelt[2] * self.units[2])
                       - convert_to_ms(0. * self.units[2])).value
            self.wcs.wcs.crval[2] = \
                convert_to_ms(self.wcs.wcs.crval[2] * self.units[2]).value
            self.wcs.wcs.ctype[2] = 'VELO-OBS'
            self.units[2] = U.m * U.s**-1
            self.wcs.wcs.cunit[2] = self.units[2].to_string('fits')
            self._channel_mids()
            self._channel_edges()
            self._freq_channel_mode = False
            return

    def add_pad(self, pad):
        """
        Resize the cube to add a padding region in the spatial direction.

        Accurate convolution with a beam requires a cube padded according to
        the size of the beam kernel (its representation sampled on a grid with
        the same spacing). The beam class is required to handle defining the
        size of pad required.

        Parameters
        ----------
        pad : tuple or other sequence of length 2
            Number of pixels to add in the x (RA) and y (Dec) directions.

        See Also
        ----------
        drop_pad
        """

        tmp = self._array
        self._array = np.zeros((self.n_px_x + pad[0] * 2,
                                self.n_px_y + pad[1] * 2, self.n_channels, 1))
        self._array = self._array * tmp.unit
        self._array[pad[0]:-pad[0], pad[1]:-pad[1], ...] = tmp
        self.wcs.wcs.crpix += np.array([pad[0], pad[1], 0, 0])
        self.padx, self.pady = pad
        return

    def drop_pad(self):
        """
        Remove the padding added using add_pad.

        After convolution, the pad region contains meaningless information and
        can be discarded.

        See Also
        --------
        add_pad
        """

        if (self.padx == 0) and (self.pady == 0):
            return
        self._array = self._array[self.padx:-self.padx, self.pady:-self.pady,
                                  ...]
        self.wcs.wcs.crpix -= np.array([self.padx, self.pady, 0, 0])
        self.padx, self.pady = 0, 0
        return

    def copy(self):
        """
        Produce a copy of the DataCube.

        May be especially useful to create multiple datacubes with differing
        intermediate steps.

        Returns
        -------
        out : DataCube
            Copy of the DataCube object.

        Examples
        --------
        TODO
        """
        copy = DataCube(self.n_px_x, self.n_px_y, self.n_channels,
                        self.px_size, self.channel_width, self.velocity_centre,
                        self.ra, self.dec)
        copy.padx, copy.pady = self.padx, self.pady
        copy.wcs = self.wcs
        copy._array = self._array.copy()
        return copy

    def __repr__(self):
        """
        Print the contents of the data cube array itself.

        Returns
        -------
        out : string
            Text representation of the DataCube._array contents.
        """
        return self._array.__repr__()



from martini import Martini, DataCube
from martini.beams import GaussianBeam
from martini.noise import GaussianNoise
from martini.spectral_models import GaussianSpectrum
from martini.sph_kernels import DiracDeltaKernel
from martini.sources import SPHSource
import astropy.units as U
import numpy as np
from scipy.optimize import fsolve


def demo(cubefile='testcube.fits', beamfile='testbeam.fits'):
    """
    Demonstrates basic usage of MARTINI.

    Creates a (very!) crude toy model of a galaxy with a linearly rising
    rotation curve, exponential disk profile, exponential vertical structure. A
    basic configuration of MARTINI is initialized and used to create and output
    a datacube and an image of the beam.

    Parameters
    ----------
    cubefile : string
        File to write demonstration datacube.

    beamfile : string
        File to write demonstration beam.
    """

    # ------make a toy galaxy----------
    N = 1000
    phi = np.random.rand(N) * 2 * np.pi
    r = []
    for L in np.random.rand(N):
        def f(r):
            return L - .5 * (2 - np.exp(-r) * (np.power(r, 2) + 2 * r + 2))
        r.append(fsolve(f, 1.)[0])
    r = np.array(r)
    # exponential disk
    r *= 3 / np.sort(r)[N // 2]
    z = -np.log(np.random.rand(N))
    # exponential scale height
    z *= .5 / np.sort(z)[N // 2] * np.sign(np.random.rand(N) - .5)
    x = r * np.cos(phi)
    y = r * np.sin(phi)
    xyz_g = np.vstack((x, y, z)) * U.kpc
    # linear rotation curve
    vphi = 100 * r / 6.
    vx = -vphi * np.sin(phi)
    vy = vphi * np.cos(phi)
    # small pure random z velocities
    vz = (np.random.rand(N) * 2. - 1.) * 5
    vxyz_g = np.vstack((vx, vy, vz)) * U.km * U.s ** -1
    T_g = np.ones(N) * 8E3 * U.K
    mHI_g = np.ones(N) / N * 5.E9 * U.Msun
    # ~mean interparticle spacing smoothing
    hsm_g = np.ones(N) * 2 / np.sqrt(N) * U.kpc
    # ---------------------------------

    source = SPHSource(
        distance=5. * U.Mpc,
        rotation={'L_coords': (60. * U.deg, 0. * U.deg)},
        ra=0. * U.deg,
        dec=0. * U.deg,
        h=.7,
        T_g=T_g,
        mHI_g=mHI_g,
        xyz_g=xyz_g,
        vxyz_g=vxyz_g,
        hsm_g=hsm_g
    )

    datacube = DataCube(
        n_px_x=128,
        n_px_y=128,
        n_channels=32,
        px_size=10. * U.arcsec,
        channel_width=10. * U.km * U.s ** -1,
        velocity_centre=source.vsys
    )

    beam = GaussianBeam(
        bmaj=30. * U.arcsec,
        bmin=30. * U.arcsec,
        bpa=0. * U.deg,
        truncate=4.
    )

    noise = GaussianNoise(
        rms=3.E-4 * U.Jy * U.arcsec ** -2
    )

    spectral_model = GaussianSpectrum(
        sigma=7 * U.km * U.s ** -1
    )

    sph_kernel = DiracDeltaKernel()

    M = Martini(
        source=source,
        datacube=datacube,
        beam=beam,
        noise=noise,
        spectral_model=spectral_model,
        sph_kernel=sph_kernel
    )

    M.insert_source_in_cube()
    M.add_noise()
    M.convolve_beam()
    M.write_beam_fits(beamfile, channels='velocity')
    M.write_fits(cubefile, channels='velocity')

    return



import numpy as np
from astropy.coordinates import CartesianRepresentation,\
    CartesianDifferential, ICRS
from astropy.coordinates.matrix_utilities import rotation_matrix
import astropy.units as U
from ._L_align import L_align
from ._cartesian_translation import translate, translate_d

# Extend CartesianRepresentation to allow coordinate translation
setattr(CartesianRepresentation, 'translate', translate)

# Extend CartesianDifferential to allow velocity (or other differential)
# translation
setattr(CartesianDifferential, 'translate', translate_d)


class SPHSource(object):
    """
    Class abstracting HI emission sources consisting of SPH simulation
    particles.

    This class constructs an HI emission source from arrays of SPH particle
    properties: mass, smoothing length, temperature, position, and velocity.

    Parameters
    ----------
    distance : astropy.units.Quantity, with dimensions of length
        Source distance, also used to set the velocity offset via Hubble's law.

    rotation : dict
        Keys may be any combination of 'axis_angle', 'rotmat' and/or
        'L_coords'. These will be applied in this order. Note that the 'y-z'
        plane will be the one eventually placed in the plane of the "sky". The
        corresponding values:
         - 'axis_angle' : 2-tuple, first element one of 'x', 'y', 'z' for the
           axis to rotate about, second element an astropy.units.Quantity with
           dimensions of angle, indicating the angle to rotate through.
         - 'rotmat' : A (3, 3) numpy.array specifying a rotation.
         - 'L_coords' : A 2-tuple containing an inclination and an azimuthal
           angle (both astropy.units.Quantity instances with dimensions of
           angle). The routine will first attempt to identify a preferred plane
           based on the angular momenta of the central 1/3 of particles in the
           source. This plane will then be rotated to lie in the plane of the
           "sky" ('y-z'), rotated by the azimuthal angle about its angular
           momentum pole (rotation about 'x'), and inclined (rotation about
           'y').

    ra : astropy.units.Quantity, with dimensions of angle
        Right ascension for the source centroid.

    dec : astropy.units.Quantity, with dimensions of angle
        Declination for the source centroid.

    h : float
        Dimensionless hubble constant, H0 = h * 100 km / s / Mpc.

    T_g : astropy.units.Quatity, with dimensions of temperature
        Particle temperature.

    mHI_g : astropy.unit.Quantity, with dimensions of mass
        Particle mass.

    xyz_g : astropy.units.Quantity array of length 3, with dimensions of length
        Particle position offset from source centroid. Note that the 'y-z'
        plane is that eventually placed in the plane of the "sky"; 'x' is
        the axis corresponding to the "line of sight".

    vxyz_g : astropy.units.Quantity array of length 3, with dimensions of
             velocity
        Particle velocity offset from source centroid. Note that the 'y-z'
        plane is that eventually placed in the plane of the "sky"; 'x' is
        the axis corresponding to the "line of sight".

    hsm_g : astropy.units.Quantity, with dimensions of length
        Particle SPH smoothing lengths.

    Returns
    -------
    out : SPHSource
        An appropriately initialized SPHSource object.

    See Also
    --------
    SingleParticleSource (simplest possible implementation of a class
                          inheriting from SPHSource).
    CrossSource
    SOSource
    """

    def __init__(
            self,
            distance=3. * U.Mpc,
            rotation={'rotmat': np.eye(3)},
            ra=0.*U.deg,
            dec=0.*U.deg,
            h=.7,
            T_g=None,
            mHI_g=None,
            xyz_g=None,
            vxyz_g=None,
            hsm_g=None,
            coordinate_axis=None
    ):

        if coordinate_axis is None:
            if (xyz_g.shape[0] == 3) and (xyz_g.shape[1] != 3):
                coordinate_axis = 0
            elif (xyz_g.shape[0] != 3) and (xyz_g.shape[1] == 3):
                coordinate_axis = 1
            elif xyz_g.shape == (3, 3):
                raise RuntimeError("martini.sources.SPHSource: cannot guess "
                                   "coordinate_axis with shape (3, 3), provide"
                                   " explicitly.")
            else:
                raise RuntimeError("martini.sources.SPHSource: incorrect "
                                   "coordinate shape (not (3, N) or (N, 3)).")

        if xyz_g.shape != vxyz_g.shape:
            raise ValueError("martini.sources.SPHSource: xyz_g and vxyz_g must"
                             " have matching shapes.")
        self.h = h
        self.T_g = T_g
        self.mHI_g = mHI_g
        self.coordinates_g = CartesianRepresentation(
            xyz_g,
            xyz_axis=coordinate_axis,
            differentials={'s': CartesianDifferential(
                vxyz_g,
                xyz_axis=coordinate_axis
            )}
        )
        self.hsm_g = hsm_g

        self.npart = self.mHI_g.size

        self.ra = ra
        self.dec = dec
        self.distance = distance
        self.rotation = rotation
        self.current_rotation = np.eye(3)
        self.rotate(**self.rotation)
        self.rotate(axis_angle=('y', self.dec))
        self.rotate(axis_angle=('z', -self.ra))
        distance_vector = np.array([
            np.cos(self.ra) * np.cos(self.dec),
            np.sin(self.ra) * np.cos(self.dec),
            np.sin(self.dec)
        ]) * self.distance
        self.translate_position(distance_vector)
        self.vsys = \
            (self.h * 100.0 * U.km * U.s ** -1 * U.Mpc ** - 1) * self.distance
        hubble_flow_vector = distance_vector * self.vsys / self.distance
        self.translate_velocity(hubble_flow_vector)
        self.sky_coordinates = ICRS(self.coordinates_g)
        return

    def apply_mask(self, mask):
        """
        Remove particles from source arrays according to a mask.

        Parameters
        ----------
        mask : array-like, containing boolean-like
            Remove particles with indices corresponding to False values from
            the source arrays.
        """

        self.T_g = self.T_g[mask]
        self.mHI_g = self.mHI_g[mask]
        self.coordinates_g = self.coordinates_g[mask]
        self.sky_coordinates = ICRS(self.coordinates_g)
        self.hsm_g = self.hsm_g[mask]
        self.npart = np.sum(mask)
        if self.npart == 0:
            raise RuntimeError('No source particles in target region.')
        return

    def rotate(self, axis_angle=None, rotmat=None, L_coords=None):
        """
        Rotate the source.

        The arguments correspond to different rotation types. If supplied
        together in one function call, they are applied in order: axis_angle,
        then rotmat, then L_coords.

        Parameters
        ----------
        axis_angle : 2-tuple
            First element one of 'x', 'y', 'z' for the axis to rotate about,
            second element an astropy.units.Quantity with dimensions of angle,
            indicating the angle to rotate through.
        rotmat : (3, 3) array-like
            Rotation matrix.
        L_coords : 2-tuple
            First element containing an inclination and second element an
            azimuthal angle (both astropy.units.Quantity instances with
            dimensions of angle). The routine will first attempt to identify
            a preferred plane based on the angular momenta of the central 1/3
            of particles in the source. This plane will then be rotated to lie
            in the 'y-z' plane, followed by a rotation by the azimuthal angle
            about its angular momentum pole (rotation about 'x'), and finally
            inclined (rotation about 'y').
        """

        do_rot = np.eye(3)

        if axis_angle is not None:
            do_rot = rotation_matrix(
                axis_angle[1],
                axis=axis_angle[0]
            ).dot(do_rot)

        if rotmat is not None:
            do_rot = rotmat.dot(do_rot)

        if L_coords is not None:
            incl, az_rot = L_coords
            do_rot = L_align(self.coordinates_g.get_xyz(),
                             self.coordinates_g.differentials['s'].get_d_xyz(),
                             self.mHI_g, frac=.3, Laxis='x').dot(do_rot)
            do_rot = rotation_matrix(az_rot, axis='x').dot(do_rot)
            do_rot = rotation_matrix(incl, axis='y').dot(do_rot)

        self.current_rotation = do_rot.dot(self.current_rotation)
        self.coordinates_g = self.coordinates_g.transform(do_rot)
        return

    def translate_position(self, translation_vector):
        """
        Translate the source.

        Note that the "line of sight" is along the 'x' axis.

        Parameters
        ----------
        translation_vector : astropy.units.Quantity, shape (3, ), with
                             dimensions of length
            Vector by which to offset the source particle coordinates.
        """

        self.coordinates_g = self.coordinates_g.translate(translation_vector)
        return

    def translate_velocity(self, translation_vector):
        """
        Apply an offset to the source velocity.

        Note that the "line of sight" is along the 'x' axis.

        Parameters
        ----------
        translation_vector : astropy.units.Quantity, shape (3, ), with
                             dimensions of velocity
            Vector by which to offset the source particle velocities.
        """

        self.coordinates_g.differentials['s'] = \
            self.coordinates_g.differentials['s'].translate(translation_vector)
        return


class SingleParticleSource(SPHSource):
    """
    Class illustrating inheritance from martini.sources.SPHSource, creates a
    single particle test source.

    A simple test source consisting of a single particle will be created. The
    particle has a mass of 10^4 Msun, a SPH smoothing length of 1 kpc, a
    temperature of 10^4 K, a position offset by (x, y, z) = (1 pc, 1 pc, 1 pc)
    from the source centroid, a peculiar velocity of 0 km/s, and will be placed
    in the Hubble flow assuming h = 0.7 and the distance provided.

    Parameters
    ----------
    distance : astropy.units.Quantity, with units of length
        Source distance, also used to place the source in the Hubble flow
        assuming h = 0.7.

    ra : astropy.units.Quantity, with dimensions of angle
        Right ascension for the source centroid.

    dec : astropy.units.Quantity, with dimensions of angle
        Declination for the source centroid.

    Returns
    -------
    out : SingleParticleSource
        An appropriately initialized SingleParticleSource object.
    """

    def __init__(self, distance=3 * U.Mpc, ra=0. * U.deg, dec=0. * U.deg):

        super().__init__(
            distance=distance,
            rotation={'rotmat': np.eye(3)},
            ra=ra,
            dec=dec,
            h=.7,
            T_g=np.ones(1) * 1.E4 * U.K,
            mHI_g=np.ones(1) * 1.E4 * U.solMass,
            xyz_g=np.ones((1, 3)) * 1.E-3 * U.kpc,
            vxyz_g=np.zeros((1, 3)) * U.km * U.s ** -1,
            hsm_g=np.ones(1) * U.kpc
            )
        return


class CrossSource(SPHSource):
    """
    Creates a source consisting of 4 particles arrayed in an asymmetric cross.

    A simple test source consisting of four particles will be created. Each has
    a mass of 10^4 Msun, a SPH smoothing length of 1 kpc, a temperature of
    10^4 K, and will be placed in the Hubble flow assuming h=.7 and a distance
    of 3 Mpc. Particle coordinates in kpc are
    [[0,  1,  0],
     [0,  0,  2],
     [0, -3,  0],
     [0,  0, -4]]
    and velocities in km/s are
    [[0,  0,  1],
     [0, -1,  0],
     [0,  0, -1],
     [0,  1,  0]]

    Parameters
    ----------
    distance : astropy.units.Quantity, with units of length
        Source distance, also used to place the source in the Hubble flow
        assuming h = 0.7.

    rotation : dict
        Keys may be any combination of 'axis_angle', 'rotmat' and/or
        'L_coords'. These will be applied in this order. Note that the 'y-z'
        plane will be the one eventually placed in the plane of the "sky". The
        corresponding values:
         - 'axis_angle' : 2-tuple, first element one of 'x', 'y', 'z' for the
           axis to rotate about, second element an astropy.units.Quantity with
           dimensions of angle, indicating the angle to rotate through.
         - 'rotmat' : A (3, 3) numpy.array specifying a rotation.
         - 'L_coords' : A 2-tuple containing an inclination and an azimuthal
           angle (both astropy.units.Quantity instances with dimensions of
           angle). The routine will first attempt to identify a preferred plane
           based on the angular momenta of the central 1/3 of particles
           in the source. This plane will then be rotated to lie in the plane
           of the "sky" ('y-z'), rotated by the azimuthal angle about its
           angular momentum pole (rotation about 'x'), and inclined (rotation
           about 'y').

    ra : astropy.units.Quantity, with dimensions of angle
        Right ascension for the source centroid.

    dec : astropy.units.Quantity, with dimensions of angle
        Declination for the source centroid.

    Returns
    -------
    out : CrossSource
        An appropriately initialized CrossSource object.
    """

    def __init__(
            self,
            distance=3. * U.Mpc,
            rotation={'rotmat': np.eye(3)},
            ra=0. * U.deg,
            dec=0. * U.deg
    ):

        xyz_g = np.array([[0, 1, 0],
                          [0, 0, 2],
                          [0, -3, 0],
                          [0, 0, -4]]) * U.kpc,

        vxyz_g = np.array([[0, 0, 1],
                           [0, -1, 0],
                           [0, 0, -1],
                           [0, 1, 0]]) * U.km * U.s ** -1

        super().__init__(
            distance=distance,
            rotation={'rotmat': np.eye(3)},
            ra=ra,
            dec=dec,
            h=.7,
            T_g=np.ones(4) * 1.E4 * U.K,
            mHI_g=np.ones(4) * 1.E4 * U.solMass,
            xyz_g=xyz_g,
            vxyz_g=vxyz_g,
            hsm_g=np.ones(4) * U.kpc
        )
        return


class SOSource(SPHSource):
    """
    Class abstracting HI sources using the SimObj package for interface to
    simulation data.

    This class accesses simulation data via the SimObj package
    (https://github.com/kyleaoman/simobj); see the documentation of that
    package for further configuration instructions.

    Parameters
    ----------
    distance : astropy.units.Quantity, with dimensions of length
        Source distance, also used to set the velocity offset via Hubble's law.

    rotation : dict
        Keys may be any combination of 'axis_angle', 'rotmat' and/or
        'L_coords'. These will be applied in this order. Note that the 'y-z'
        plane will be the one eventually placed in the plane of the "sky". The
        corresponding values:
         - 'axis_angle' : 2-tuple, first element one of 'x', 'y', 'z' for the
           axis to rotate about, second element an astropy.units.Quantity with
           dimensions of angle, indicating the angle to rotate through.
         - 'rotmat' : A (3, 3) numpy.array specifying a rotation.
         - 'L_coords' : A 2-tuple containing an inclination and an azimuthal
           angle (both astropy.units.Quantity instances with dimensions of
           angle). The routine will first attempt to identify a preferred plane
           based on the angular momenta of the central 1/3 of particles in the
           source. This plane will then be rotated to lie in the plane of the
           "sky" ('y-z'), rotated by the azimuthal angle about its angular
           momentum pole (rotation about 'x'), and inclined (rotation about
           'y').

    ra : astropy.units.Quantity, with dimensions of angle
        Right ascension for the source centroid.

    dec : astropy.units.Quantity, with dimensions of angle
        Declination for the source centroid.

    SO_args : dict
        Dictionary of keyword arguments to pass to a call to simobj.SimObj.
        Arguments are: 'obj_id', 'snap_id', 'mask_type', 'mask_args',
        'mask_kwargs', 'configfile', 'simfiles_configfile', 'ncpu'. See simobj
        package documentation for details.

    Returns
    -------
    out : SOSource
        An appropriately initialized SOSource object.
    """

    def __init__(
            self,
            distance=3.*U.Mpc,
            rotation={'L_coords': (60.*U.deg, 0.*U.deg)},
            ra=0.*U.deg,
            dec=0.*U.deg, SO_args=dict()
    ):

        from simobj import SimObj  # optional dependency for this source class
        
        self._SO_args = SO_args
        with SimObj(**self._SO_args) as SO:
            super().__init__(
                distance=distance,
                rotation=rotation,
                ra=ra,
                dec=dec,
                h=SO.h,
                T_g=SO.T_g,
                mHI_g=SO.mHI_g,
                xyz_g=SO.xyz_g,
                vxyz_g=SO.vxyz_g,
                hsm_g=SO.hsm_g
                )
        return



from abc import ABCMeta, abstractmethod
import numpy as np
import astropy.units as U


class _BaseNoise(object):
    """
    Abstract base class to inherit for classes implementing a noise model.

    Classes inheriting from _BaseNoise must implement a method 'generate' which
    receives one argument, a martini.DataCube instance. This method should
    return an astropy.units.Quantity array with the same shape as the datacube
    array, and units of Jy (or compatible).

    See Also
    --------
    GaussianNoise (simple example implementing a Gaussian noise model)
    """

    __metaclass__ = ABCMeta

    def __init__(self):
        return

    @abstractmethod
    def generate(self, datacube):
        """
        Abstract method; create a cube containing noise.
        """
        pass


class GaussianNoise(_BaseNoise):
    """
    Implementation of a simple Gaussian noise model.

    Provides a 'generate' method producing a cube of Gaussian Noise.

    Parameters
    ----------
    rms : astropy.units.Quantity with dimensions of flux density (e.g. Jy)
        Root mean square amplitude of the noise field.

    Returns
    -------
    out : GaussianNoise
        An appropriately initialized GaussianNoise object.
    """

    def __init__(self, rms=1.0 * U.Jy):

        self.rms = rms

        super().__init__()

        return

    def generate(self, datacube):
        """
        Create a cube containing Gaussian noise.

        Some numpy functions such as np.random.normal strip units, so need to
        handle them explicitly.

        Parameters
        ----------
        datacube : martini.DataCube instance
            This method will be called passing the martini.DataCube instance as
            an argument; its attributes can thus be accessed here.
            datacube._array.shape is particularly relevant.

        Returns
        -------
        out : astropy.units.Quantity array, with dimensions of flux density
            Noise realization with size matching the DataCube._array.
        """

        return np.random.normal(
            scale=self.rms.value,
            size=datacube._array.shape
        ) * self.rms.unit



from setuptools import setup

setup(
    name='astromartini',
    version='1.0.0',
    description='Synthetic datacube creation from simulations.',
    url='',
    author='Kyle Oman',
    author_email='koman@astro.rug.nl',
    license='',
    packages=['martini'],
    install_requires=['numpy', 'astropy', 'scipy'],
    include_package_data=True,
    zip_safe=False
)



import numpy as np

# copied from github.com/kyleaoman/kyleaoman_utilities/kyleaoman_utilities/
# commit-id 81e08768bcf3f910d86757c07b44632f393f29aa
# Note: No git-based solution (e.g. via submodules) seems practical to include
# selected files from external repositories; a direct copy is included here
# to produce a self-contained package.


def L_align(xyz, vxyz, m, frac=.3, saverot=None, Laxis='z'):

    transposed = False
    if xyz.ndim != 2:
        raise ValueError('L_align: cannot guess coordinate axis for input with'
                         ' ndim != 2.')
    elif (xyz.shape[0] == 3) and (xyz.shape[1] == 3):
        raise ValueError('L_align: cannot guess coordinate axis for input with'
                         ' shape (3, 3).')
    elif xyz.shape[1] == 3:
        xyz = xyz.T
        vxyz = vxyz.T
        transposed = True
    elif (xyz.shape[0] != 3) and (xyz.shape[1] != 3):
        raise ValueError('L_align: coordinate array shape '+str(xyz.shape) +
                         ' invalid (one dim must be 3).')

    rsort = np.argsort(np.sum(np.power(xyz, 2), axis=0), kind='quicksort')
    p = m[np.newaxis] * vxyz
    L = np.cross(xyz, p, axis=0)
    p = p[:, rsort]
    L = L[:, rsort]
    m = m[rsort]
    mcumul = np.cumsum(m) / np.sum(m)
    Nfrac = np.argmin(np.abs(mcumul - frac))
    Nfrac = np.max([Nfrac, 100])  # use a minimum of 100 particles
    Nfrac = np.min([Nfrac, len(m)])  # unless this exceeds particle count
    p = p[:Nfrac]
    L = L[:Nfrac]
    Ltot = np.sqrt(np.sum(np.power(np.sum(L, axis=1), 2)))
    Lhat = np.sum(L, axis=1) / Ltot
    zhat = Lhat / np.sqrt(np.sum(np.power(Lhat, 2)))  # normalized
    xaxis = np.array([1., 1., 1.])  # default unlikely Laxis
    xhat = xaxis - xaxis.dot(zhat) * zhat
    xhat = xhat / np.sqrt(np.sum(np.power(xhat, 2)))  # normalized
    yhat = np.cross(zhat, xhat)  # guarantees right-handedness

    rotmat = np.vstack((xhat, yhat, zhat))  # units will be dropped (desired)
    if Laxis == 'z':
        pass
    elif Laxis == 'y':
        rotmat = np.roll(rotmat, 2, axis=0)
    elif Laxis == 'x':
        rotmat = np.roll(rotmat, 1, axis=0)
    else:
        raise ValueError("L_align: Laxis must be one of 'x', 'y' or 'z'.")

    if transposed:
        rotmat = rotmat.T
    if saverot is not None:
        np.save(saverot, rotmat)

    return rotmat



from abc import ABCMeta, abstractmethod
import numpy as np
import astropy.units as U
from astropy.io import fits
from astropy import wcs
from scipy.interpolate import RectBivariateSpline
import warnings
import os.path

f_HI = 1.420405751 * U.GHz


class _BaseBeam(object):
    """
    Abstract base class for classes implementing a radio telescope beam model.

    Classes inheriting from _BaseBeam must implement three methods: 'f_kernel',
    'kernel_size_px' and 'init_beam_header'.

    'f_kernel' should return a function accepting two arguments, the RA and Dec
    offsets from the beam centroid (provided with units of arcsec), and
    returning the beam amplitude at that location.

    'kernel_px_size' should return a 2-tuple containing the half-size (x, y) of
    the beam image that will be initialized, in pixels.

    'init_beam_header' should be defined if the major/minor axis FWHM of the
    beam and its position angle are not defined when the beam object is
    initialized, for instance if modelling a particular telescope this function
    can be used to set the (constant) parameters of the beam of that particular
    facility.

    Parameters
    ----------
    bmaj : astropy.units.Quantity, with dimensions of angle
        Beam major axis (FWHM) angular size.

    bmin : astropy.units.Quantity, with dimensions of angle
        Beam minor axis (FWHM) angular size.

    bpa : astropy.units.Quantity, with dimesions of angle
        Beam position angle (East from North).

    See Also
    --------
    GaussianBeam
    """

    __metaclass__ = ABCMeta

    def __init__(self,
                 bmaj=15. * U.arcsec,
                 bmin=15. * U.arcsec,
                 bpa=0. * U.deg):
        # some beams need information from the datacube; in this case make
        # their call to _BaseBeam.__init__ with bmaj == bmin == bpa == None
        # and define a init_beam_header, to be called after the ra, dec,
        # vel, etc. of the datacube are known
        self.bmaj = bmaj
        self.bmin = bmin
        self.bpa = bpa
        self.px_size = None
        self.kernel = None

        return

    def needs_pad(self):
        """
        Determine the padding of the datacube required by the beam to prevent
        edge effects during convolution.

        Returns
        -------
        2-tuple, each element an integer
        """

        return self.kernel.shape[0] // 2, self.kernel.shape[1] // 2

    def init_kernel(self, datacube):
        """
        Calculate the required size of the beam image

        Parameters
        ----------
        datacube : DataCube instance
            Datacube to use, cube size is required for pixel size, position &
            velocity centroids.
        """

        self.px_size = datacube.px_size
        self.vel = datacube.velocity_centre
        self.ra = datacube.ra
        self.dec = datacube.dec
        if (self.bmaj is None) or (self.bmin is None) or (self.bpa is None):
            self.init_beam_header()
        npx_x, npx_y = self.kernel_size_px()
        px_centres_x = (np.arange(-npx_x, npx_x + 1)) * self.px_size
        px_centres_y = (np.arange(-npx_y, npx_y + 1)) * self.px_size
        self.kernel = self.f_kernel()(*np.meshgrid(px_centres_x, px_centres_y))
        self.arcsec_to_beam = (U.Jy * U.arcsec**-2, U.Jy * U.beam**-1,
                               lambda x: x * (np.pi * self.bmaj * self.bmin),
                               lambda x: x / (np.pi * self.bmaj * self.bmin))

        # can turn 2D beam into a 3D beam here; use above for central channel
        # then shift in frequency up and down for other channels
        # then probably need to adjust convolution step to do the 2D
        # convolution on a stack

        return

    @abstractmethod
    def f_kernel(self):
        """
        Abstract method; returns a function defining the beam amplitude as a
        function of position.

        The function returned by this method should accept two parameters, the
        RA and Dec offset from the beam centroid, and return the beam amplitude
        at that position. The offsets are provided as astropy.units.Quantity
        objects with dimensions of angle (arcsec).
        """

        pass

    @abstractmethod
    def kernel_size_px(self):
        """
        Abstract method; returns a 2-tuple specifying the half-size of the beam
        image to be initialized, in pixels.
        """

        pass

    @abstractmethod
    def init_beam_header(self):
        """
        Abstract method; sets beam major/minor axis lengths and position angle.

        This method is optional, and only needs to be defined if these
        parameters are not specified in the call to the __init__ method of the
        derived class.
        """

        pass


class GaussianBeam(_BaseBeam):
    """
    Class implementing a Gaussiam beam model.

    Paramters
    ---------
    bmaj : astropy.units.Quantity, with dimensions of angle
        Beam major axis (FWHM) angular size.

    bmin : astropy.units.Quantity, with dimensions of angle
        Beam minor axis (FWHM) angular size.

    bpa : astropy.units.Quantity, with dimensions of angle
        Beam position angle (East of North).

    truncate : float
        Number of FWHM at which to truncate the beam image.
    """

    def __init__(self,
                 bmaj=15. * U.arcsec,
                 bmin=15. * U.arcsec,
                 bpa=0. * U.deg,
                 truncate=4.):
        self.truncate = truncate
        super().__init__(bmaj=bmaj, bmin=bmin, bpa=bpa)
        return

    def f_kernel(self):
        """
        Returns a function defining the beam amplitude as a function of
        position.

        The model implemented is a 2D Gaussian with FWHM's specified by bmaj
        and bmin and orientation by bpa.

        Returns
        -------
        Callable accepting 2 arguments (both float or array) and returning a
        float or array of corresponding size.
        """

        def fwhm_to_sigma(fwhm):
            return fwhm / (2. * np.sqrt(2. * np.log(2.)))

        sigmamaj = fwhm_to_sigma(self.bmaj)
        sigmamin = fwhm_to_sigma(self.bmin)

        a = np.power(np.cos(self.bpa), 2) / (2. * np.power(sigmamin, 2)) \
            + np.power(np.sin(self.bpa), 2) / (2. * np.power(sigmamaj, 2))
        # signs set for CCW rotation (PA)
        b = -np.sin(2. * self.bpa) / (4 * np.power(sigmamin, 2)) \
            + np.sin(2. * self.bpa) / (4 * np.power(sigmamaj, 2))
        c = np.power(np.sin(self.bpa), 2) / (2. * np.power(sigmamin, 2)) \
            + np.power(np.cos(self.bpa), 2) / (2. * np.power(sigmamaj, 2))
        A = np.power(2. * np.pi * sigmamin * sigmamaj, -1)
        A *= np.power(
            self.px_size, 2
        ).value
        # above causes an extra factor of pixel area, need to track this down
        # properly an see whether correction should apply to all beams, or
        # somewhere else?

        return lambda x, y: A * np.exp(
            -a * np.power(x, 2) - 2. * b * x * y - c * np.power(y, 2)
        )

    def kernel_size_px(self):
        """
        Returns a 2-tuple specifying the half-size of the beam image to be
        initialized, in pixels.

        Returns
        -------
        2-tuple, each element an integer.
        """

        size = np.ceil(
            (self.bmaj * self.truncate)
            .to(U.pix, U.pixel_scale(self.px_size / U.pix))
        ).value + 1
        return size, size


class WSRTBeam(_BaseBeam):
    """
    Class implementing a beam model for the Westerbork Synthesis Radio
    Telescope.

    The beam is adapted from an actual image from the telescope. The image is
    scaled for frequency and declination then interpolated to the required
    resolution. Note that the image is for the "classic" WSRT, not APERTIF.

    The input image is located at:
    '<martini directory>/data/beam00_freq02.fits'
    """

    beamfile = os.path.join(
        os.path.dirname(__file__), 'data/beam00_freq02.fits')

    def __init__(self):
        super().__init__(bmaj=None, bmin=None, bpa=None)

    def _load_beamfile(self):
        """
        Open the input image.

        Returns
        -------
        2-tuple containing an astropy.io.fits.Header and an
        astropy.units.Quantity array.
        """
        with fits.open(self.beamfile) as f:
            return f[0].header, np.transpose(
                f[0].data, axes=(1, 2, 0)) * U.Jy / U.beam

    def _centroid(self):
        """
        Determine the centre in RA, Dec and frequency of the input image.

        Returns
        -------
        3-tuple containing: three astropy.units.Quantity objects with
        dimensions of angle, angle and frequency, respectively, corresponding
        to the RA, Dec and frequency centres.
        """

        bheader, bdata = self._load_beamfile()
        c = wcs.WCS(header=bheader)
        RAgrid, Decgrid, freqgrid = c.wcs_pix2world(
            *np.meshgrid(*tuple([np.arange(N) for N in bdata.shape])), 0)
        RAgrid *= U.deg
        Decgrid *= U.deg
        freqgrid *= U.Hz
        return tuple([
            A[tuple(np.array(bdata.shape) // 2)]
            for A in (RAgrid, Decgrid, freqgrid)
        ])

    def init_beam_header(self):
        """
        Set beam major/minor axis lengths and position angle to WSRT values.
        """

        self.bmaj = 15. * U.arcsec / np.sin(self.dec)
        self.bmin = 15. * U.arcsec
        self.bpa = 90. * U.deg
        return

    def f_kernel(self):
        """
        Returns a function defining the beam amplitude as a function of
        position.

        The model implemented is based on an interpolation of an image of the
        WSRT beam, re-scaled according to declination and frequency.

        Returns
        -------
        Callable accepting 2 arguments (both float or array) and returning a
        float or array of corresponding size.
        """

        if self.dec <= 0. * U.deg:
            raise ValueError('WSRT beam requires positive declination.')
        freq = self.vel.to(U.GHz, equivalencies=U.doppler_radio(f_HI))
        bheader, bdata = self._load_beamfile()
        centroid = self._centroid()
        bpx_ra = np.abs(bheader['CDELT1'] * U.deg).to(U.arcsec)
        bpx_dec = np.abs(bheader['CDELT2'] * U.deg).to(U.arcsec)
        dRAs = np.arange(-(bdata.shape[0] // 2), bdata.shape[0] // 2 + 1) \
            * bpx_ra * (centroid[2] / freq).to(U.dimensionless_unscaled)
        dDecs = np.arange(-(bdata.shape[1] // 2), bdata.shape[1] // 2 + 1) \
            * bpx_dec * np.sin(self.dec) / np.sin(centroid[1]) \
            * (centroid[2] / freq)\
            .to(U.dimensionless_unscaled)
        interpolator = RectBivariateSpline(
            dRAs, dDecs, bdata[..., 0], kx=1, ky=1, s=0)
        # RectBivariateSpline is a wrapper around Fortran code and causes a
        # transpose
        return lambda x, y: interpolator(y, x, grid=False).T

    def kernel_size_px(self):
        """
        Returns a 2-tuple specifying the half-size of the beam image to be
        initialized, in pixels.

        The interpolation may become unstable if the image is severely
        undersampled, pixel sizes of more than 8 arcsec are not recommended.

        Returns
        -------
        2-tuple, each element an integer.
        """

        if self.px_size > 12. * U.arcsec:
            warnings.warn(
                "Using WSRT beam with datacube pixel size >> 8 arcsec,"
                " beam interpolation may fail.")
        freq = self.vel.to(U.GHz, equivalencies=U.doppler_radio(f_HI))
        bheader, bdata = self._load_beamfile()
        centroid = self._centroid()
        aspect_x, aspect_y = np.floor(
            bdata.shape[0] // 2 * np.sin(self.dec)), bdata.shape[1] // 2
        aspect_x = aspect_x * np.abs((bheader['CDELT1'] * U.deg)).to(U.arcsec)\
            * (centroid[2] / freq).to(U.dimensionless_unscaled)
        aspect_y = aspect_y * (bheader['CDELT2'] * U.deg).to(U.arcsec) \
            * (centroid[2] / freq).to(U.dimensionless_unscaled)
        return tuple([
            (a.to(U.pix, U.pixel_scale(self.px_size / U.pix))).value + 1
            for a in (aspect_x, aspect_y)
        ])



from abc import ABCMeta, abstractmethod
import numpy as np
import astropy.units as U


class _BaseSPHKernel(object):
    """
    Abstract base class for classes implementing SPH kernels to inherit from.

    Classes inheriting from _BaseSPHKernel must implement two methods:
    'px_weight' and 'validate'.

    'px_weight' should define the integral of the kernel over a pixel given the
    distance between the pixel centre and the particle centre, and the
    smoothing length (both in units of pixels). The integral should be
    normalized so that evaluated over the entire kernel it is equal to 1.

    'validate' should check whether any approximations converge to sufficient
    accuracy (for instance, depending on the ratio of the pixel size and
    smoothing length), and raise an error if not.
    """

    __metaclass__ = ABCMeta

    def __init__(self):
        return

    @abstractmethod
    def px_weight(self, dij, h):
        """
        Abstract method; calculate the kernel integral over a pixel.

        Parameters
        ----------
        dij : astropy.units.Quantity, with dimensions of pixels
            Distances from pixel centre to particle positions, in pixels.

        h : astropy.units.Quantity, with dimensions of pixels
            Particle smoothing lengths, in pixels.
        """

        pass

    @abstractmethod
    def validate(self, sm_lengths):
        """
        Abstract method; check conditions for validity of kernel integral
        calculation.

        Some approximations may only converge if the ratio of the pixel size
        and the smoothing length is sufficiently large, or sufficiently small.
        This method should check these conditions and raise errors when
        appropriate. The smoothing lengths are provided normalized to the pixel
        size.

        Parameters
        ----------
        sm_lengths : astropy.units.Quantity, with dimensions of pixels
            Particle smoothing lengths, in units of pixels.
        """

        pass


class WendlandC2Kernel(_BaseSPHKernel):
    """
    Implementation of the Wendland C2 kernel integral.

    The Wendland C2 kernel is used in the EAGLE code and derivatives (not in
    Gadget/Gadget2!). The exact integral is usually too slow to be practical;
    the implementation here approximates the kernel amplitude as constant
    across the pixel, which converges to within 1% of the exact integral
    provided the SPH smoothing lengths are at least 2 pixels in size.

    Returns
    -------
    out : WendlandC2Kernel
        An appropriately initialized WendlandC2Kernel object.
    """
    def __init__(self):
        super().__init__()
        return

    def px_weight(self, dij, h):
        """
        Calculate the kernel integral over a pixel. The formula used
        approximates the kernel amplitude as constant across the pixel area and
        converges to the true value within 1% for smoothing lengths >= 2
        pixels.

        Parameters
        ----------
        dij : astropy.units.Quantity, with dimensions of pixels
            Distances from pixel centre to particle positions, in pixels.

        h : astropy.units.Quantity, with dimensions of pixels
            Particle smoothing lengths, in pixels.

        Returns
        -------
        out : np.array
            Approximate kernel integral over the pixel area.
        """

        dr2 = np.power(dij, 2).sum(axis=0)
        retval = np.zeros(h.shape)
        R2 = dr2 / (h * h)
        retval[R2 == 0] = 2. / 3.
        use = np.logical_and(R2 < 1, R2 != 0)
        R2 = R2[use]
        A = np.sqrt(1 - R2)
        retval[use] = 5 * R2 * R2 * (.5 * R2 + 3) * \
            np.log((1 + A) / np.sqrt(R2)) + \
            A * (-27. / 2. * R2 * R2 - 14. / 3. * R2 + 2. / 3.)
        # .2992 is normalization s.t. kernel integral = 1 for particle mass = 1
        return retval / .2992 / np.power(h, 2)

    def validate(self, sm_lengths):
        """
        Check conditions for validity of kernel integral calculation.

        Convergence within 1% of the exact integral is achieved when the
        smoothing lengths are >= 2 pixels.

        Parameters
        ----------
        sm_lengths : astropy.units.Quantity, with dimensions of pixels
            Particle smoothing lengths, in units of pixels.x
        """

        if (sm_lengths < 2 * U.pix).any():
            raise RuntimeError("Martini.sph_kernels.WendlandC2Kernel.validate:"
                               " SPH smoothing lengths must be >= 2 px in size"
                               " for WendlandC2 kernel integral approximation "
                               "accuracy.")
        return


class DiracDeltaKernel(_BaseSPHKernel):
    """
    Implementation of a Dirac-delta kernel integral.

    Returns
    -------
    out : DiracDeltaKernel
        An appropriately initialized DiracDeltaKernel object.
    """

    def __init__(self):
        super().__init__()
        return

    def px_weight(self, dij, h):
        """
        Calculate the kernel integral over a pixel. The particles are
        approximated as point-like, ignoring any finite-sized kernel.
        This is a reasonable approximation provided the smoothing
        length is < 0.5 pixel in size, ideally << 1 pixel in size.

        Parameters
        ----------
        dij : astropy.units.Quantity, with dimensions of pixels
            Distances from pixel centre to particle positions, in pixels.

        h : astropy.units.Quantity, with dimensions of pixels
            Particle smoothing lengths, in pixels.

        Returns
        -------
        out : np.array
            Kernel integral over the pixel area.
        """

        return np.where((np.abs(dij) < 0.5 * U.pix).all(axis=0), 1, 0) \
            * U.pix ** -2

    def validate(self, sm_lengths):

        """
        Check conditions for validity of kernel integral calculation.

        The Dirac-delta model approaches the exact integral when the smoothing
        length is << 1 pixel in size; at a minimum the smoothing length should
        be less than half the pixel size.

        Parameters
        ----------
        sm_lengths : astropy.units.Quantity, with dimensions of pixels
            Particle smoothing lengths, in units of pixels.
        """

        if (sm_lengths > .5 * U.pix).any():
            raise RuntimeError("Martini.sph_kernels.DiracDeltaKernel.validate:"
                               " SPH smoothing lengths must be <= 1 px in size"
                               " for DiracDelta kernel to be a reasonable "
                               "approximation. Initialize with "
                               "'ignore_smoothing=True' to override.")
        return



import numpy as np
import astropy.units as U
from astropy import constants as C
from scipy.special import erf as _erf
from abc import ABCMeta, abstractmethod


def erf(z):
    return _erf(z.to(U.dimensionless_unscaled).value)


class _BaseSpectrum(object):
    """
    Abstract base class for implementions of spectral models to inherit from.

    Classes inheriting from _BaseSpectrum must implement three methods:
    'half_width', 'spectral_function' and 'spectral_function_kwargs'.

    'half_width' should define a characteristic width for the model, measured
    from the peak to the characteristic location. Note that particles whose
    spectra within +/- 4 half-widths of the peak do not intersect the DataCube
    bandpass will be discarded to speed computation.

    'spectral_function' should define the model spectrum. Arguments which
    depend on the martini.sources.SPHSource (or derived class) properties
    should make use of the 'spectral_function_kwargs' method. The spectrum
    should integrate to 1, the amplitude is handled separately.

    'spectral_function_kwargs' should provide a helper to pass properties from
    martini.source.SPHSource (or derived class) to the 'spectral_function'.
    This is required because the source object is not accessible at class
    initialization.

    See Also
    --------
    GaussianSpectrum (simple example of a derived class using source properties
                      in spectral_function)
    DiracDeltaSpectrum
    """

    __metaclass__ = ABCMeta

    def __init__(self):
        self.spectra = None
        return

    def init_spectra(self, source, datacube):
        """
        Pre-compute the spectrum of each particle.

        The spectral model defined in 'spectral_function' is evaluated using
        the channel edges from the DataCube instance and the particle
        velocities of the SPHSource (or derived class) instance provided.
        Additional particle properties can be accessed via the
        'spectral_function_kwargs' helper method.

        Parameters
        ----------
        source : martini.sources.SPHSource (or derived class) instance
            Source object containing arrays of particle properties.

        datacube : martini.DataCube instance
            DataCube object defining the observational parameters, including
            spectral channels.
        """

        channel_edges = datacube.channel_edges
        channel_widths = np.diff(channel_edges).to(U.km * U.s ** -1)
        vmids = source.sky_coordinates.radial_velocity
        A = source.mHI_g * \
            np.power(source.sky_coordinates.distance.to(U.Mpc), -2)
        MHI_Jy = (
            U.solMass * U.Mpc ** -2 * (U.km * U.s ** -1) ** -1,
            U.Jy,
            lambda x: (1 / 2.36E5) * x,
            lambda x: 2.36E5 * x
        )
        spectral_function_kwargs = \
            {k: np.tile(v, np.shape(channel_edges[:-1]) + (1,) * vmids.ndim).T
             for k, v in self.spectral_function_kwargs(source).items()}
        raw_spectra = self.spectral_function(
            np.tile(
                channel_edges.value[:-1],
                vmids.shape + (1,)
            ) * channel_edges.unit,
            np.tile(
                channel_edges.value[1:],
                vmids.shape + (1,)
            ) * channel_edges.unit,
            np.tile(
                vmids.value,
                np.shape(channel_edges[:-1]) + (1,) * vmids.ndim
            ).T * vmids.unit,
            **spectral_function_kwargs
        ).to(U.dimensionless_unscaled).value
        self.spectra = (A[..., np.newaxis] * raw_spectra / channel_widths)\
            .to(U.Jy, equivalencies=[MHI_Jy])

        return

    @abstractmethod
    def half_width(self, source):
        """
        Abstract method; calculate the half-width of the spectrum, either
        globally or per-particle.

        Parameters
        ----------
        source : martini.sources.SPHSource (or derived class) instance
            The source object will be provided to allow access to particle
            properties.
        """
        pass

    @abstractmethod
    def spectral_function(self, a, b, vmids, **kwargs):
        """
        Abstract method; implementation of the spectral model.

        Should calculate the flux in each spectral channel, calculation should
        be vectorized (numpy).

        Parameters
        ----------
        a : astropy.units.Quantity array, with dimensions of velocity
            Lower spectral channel edge(s).

        b : astropy.units.Quantity array, with dimensions of velocity
            Upper spectral channel edge(s).

        vmids : astropy.units.Quantity array, with dimensions of velocity
            Particle velocities along the line of sight.

        **kwargs : various
            See spectral_function_kwargs.

        See Also
        --------
        spectral_function_kwargs
        """

        pass

    @abstractmethod
    def spectral_function_kwargs(self, source):
        """
        Abstract method; helper method to pass additional arguments to the
        spectral_function.

        Should return a dict containing the kwarg names as keys with the values
        to pass as associated values.

        Parameters
        ----------
        source : martini.sources.SPHSource (or derived class) instance
            The source object will be provided so that its attributes can be
            accessed.

        See Also
        --------
        GaussianSpectrum (example implementation)
        """

        pass


class GaussianSpectrum(_BaseSpectrum):
    """
    Class implementing a Gaussian model for the spectrum of the HI line.

    The line is modelled as a Gaussian of either fixed width, or of width
    scaling with the particle temperature as sqrt(k_B * T / m_p), centered
    at the particle velocity.

    Parameters
    ----------
    sigma : astropy.units.Quantity, with dimensions of velocity, or string:
            'thermal'
        Width of the Gaussian modelling the line (constant for all particles),
        or specify 'thermal' for width equal to sqrt(k_B * T / m_p) where k_B
        is Boltzmann's constant, T is the particle temperature and m_p is the
        particle mass.

    Returns
    -------
    out : GaussianSpectrum
        An appropriately initialized GaussianSpectrum object.

    See Also
    --------
    _BaseSpectrum
    DiracDeltaSpectrum
    """

    def __init__(self, sigma=7. * U.km * U.s ** -1):

        self.sigma_mode = sigma
        super().__init__()

        return

    def spectral_function(self, a, b, vmids, sigma=1.0 * U.km * U.s ** -1):
        """
        Evaluate a Gaussian integral in a channel.

        Parameters
        ----------
        a : astropy.units.Quantity, with dimensions of velocity
            Lower spectral channel edge(s).

        b : astropy.units.Quantity, with dimensions of velocity
            Upper spectral channel edge(s).

        vmids : astropy.units.Quantity array, with dimensions of velocity
            Particle velocities along the line of sight.

        sigma : astropy.units.Quantity array, or astropy.units.Quantity, with
                dimensions of velocity
            Velocity dispersion for HI line width, either for each particle or
            constant.

        Returns
        -------
        out : astropy.units.Quantity, dimesionless
            The evaluated spectral model.
        """

        # erf strips units
        return .5 * (
            erf((b - vmids) / (np.sqrt(2.) * sigma)) -
            erf((a - vmids) / (np.sqrt(2.) * sigma))
        ) * U.dimensionless_unscaled

    def spectral_function_kwargs(self, source):
        """
        Helper function to pass particle velocity dispersions to the
        spectral_function.

        Parameters
        ----------
        source : martini.sources.SPHSource (or derived class) instance
            Source object.

        Returns
        -------
        out : dict
            Keyword arguments for the spectral_function.
        """

        return {'sigma': self.half_width(source)}

    def half_width(self, source):
        """
        Calculate velocity dispersions from particle temperatures, or return
        constant.

        Parameters
        ----------
        source : martini.sources.SPHSource (or derived class) instance
            Source object, making particle properties available.

        Returns
        -------
        out : astropy.unit.Quantity, with dimensions of velocity
            Velocity dispersion (constant, or per particle).
        """

        if self.sigma_mode == 'thermal':
            return np.sqrt(C.k_B * source.T_g / C.m_p).to(U.km * U.s ** -1)
        else:
            return self.sigma_mode


class DiracDeltaSpectrum(_BaseSpectrum):
    """
    Class implemeting a Dirac-delta model for the spectrum of the HI line.

    The line is modelled as a Dirac-delta function, centered at the particle
    velocity.

    Returns
    -------
    out : DiracDeltaSpectrum
        An appropriately initialized DiracDeltaSpectrum object.
    """

    def __init__(self):
        super().__init__()
        return

    def spectral_function(self, a, b, vmids):
        """
        Evaluate a Dirac-delta function in a channel.

        Parameters
        ----------
        a : astropy.units.Quantity, with dimensions of velocity
            Lower spectral channel edge(s).

        b : astropy.units.Quantity, with dimensions of velocity
            Upper spectral channel edge(s).

        vmids : astropy.units.Quantity array, with dimensions of velocity
            Particle velocities along the line of sight.

        Returns
        -------
        out : astropy.units.Quantity, dimesionless
            The evaluated spectral model.
        """

        return np.heaviside(vmids - a, 1.) * np.heaviside(b - vmids, 0.)

    def spectral_function_kwargs(self, source):
        """
        No additional kwargs.

        Parameters
        ----------
        source : martini.sources.SPHSource (or derived class) instance
            Source object, making particle properties available.

        Returns
        -------
        out : dict
            Empty; no additional kwargs.
        """

        return dict()

    def half_width(self, source):
        """
        Dirac-delta function has 0 width.

        Parameters
        ----------
        source : martini.sources.SPHSource (or derived class) instance
            Source object, making particle properties available.

        Returns
        -------
        out : astropy.units.Quantity
            Velocity dispersion of 0 km/s.
        """

        return 0 * U.km * U.s ** -1
